@@ -11,6 +11,7 @@ import signal
 import subprocess
 import sys
 import tempfile
+import threading
 import time
 import unicodedata
 from six.moves import range
@@ -33,7 +34,7 @@ from bpython.config import (Struct, loadini, default_config_path,
 from bpython.formatter import BPythonFormatter
 from bpython import autocomplete
 from bpython.translations import _
-from bpython._py3compat import py3
+from bpython._py3compat import py3, is_main_thread
 from bpython.pager import get_pager_command
 
 from bpython.curtsiesfrontend import replpainter as paint
@@ -128,7 +129,7 @@ class FakeStdin(object):
             self.cursor_offset, self.current_line = self.rl_char_sequences[e](
                 self.cursor_offset, self.current_line)
         elif isinstance(e, events.SigIntEvent):
-            self.coderunner.sigint_happened_in_main_greenlet = True
+            self.coderunner.sigint_happened_in_main_context = True
             self.has_focus = False
             self.current_line = ''
             self.cursor_offset = 0
@@ -179,7 +180,7 @@ class FakeStdin(object):
     def readline(self):
         self.has_focus = True
         self.repl.send_to_stdin(self.current_line)
-        value = self.coderunner.request_from_main_greenlet()
+        value = self.coderunner.request_from_main_context()
         self.readline_results.append(value)
         return value
 
@@ -228,7 +229,7 @@ class ReevaluateFakeStdin(object):
             value = self.readline_results.pop(0)
         else:
             value = 'no saved input available'
-        self.repl.send_to_stdout(value)
+        self.repl.send_to_stdouterr(value)
         return value
 
 
@@ -332,7 +333,7 @@ class BaseRepl(BpythonRepl):
 
         if interp is None:
             interp = Interp(locals=locals_)
-            interp.write = self.send_to_stderr
+            interp.write = self.send_to_stdouterr
         if banner is None:
             if config.help_key:
                 banner = (_('Welcome to bpython!') + ' ' +
@@ -392,9 +393,9 @@ class BaseRepl(BpythonRepl):
 
         # filenos match the backing device for libs that expect it,
         # but writing to them will do weird things to the display
-        self.stdout = FakeOutput(self.coderunner, self.send_to_stdout,
+        self.stdout = FakeOutput(self.coderunner, self.send_to_stdouterr,
                                  fileno=sys.__stdout__.fileno())
-        self.stderr = FakeOutput(self.coderunner, self.send_to_stderr,
+        self.stderr = FakeOutput(self.coderunner, self.send_to_stdouterr,
                                  fileno=sys.__stderr__.fileno())
         self.stdin = FakeStdin(self.coderunner, self, self.edit_keys)
 
@@ -403,7 +404,7 @@ class BaseRepl(BpythonRepl):
 
         self.request_paint_to_pad_bottom = 0
 
-        # offscreen command yields results different from scrollback bufffer
+        # offscreen command yields results different from scrollback buffer
         self.inconsistent_history = False
 
         # history error message has already been displayed
@@ -420,7 +421,8 @@ class BaseRepl(BpythonRepl):
         self.paste_mode = False
         self.current_match = None
         self.list_win_visible = False
-        self.watching_files = False    # whether auto reloading active
+        # whether auto reloading active
+        self.watching_files = config.default_autoreload
 
         # 'reverse_incremental_search', 'incremental_search' or None
         self.incr_search_mode = None
@@ -435,6 +437,8 @@ class BaseRepl(BpythonRepl):
         self.status_bar.message(banner)
 
         self.watcher = ModuleChangedEventHandler([], self.request_reload)
+        if self.watcher and config.default_autoreload:
+            self.watcher.activate()
 
     # The methods below should be overridden, but the default implementations
     # below can be used as well.
@@ -483,13 +487,13 @@ class BaseRepl(BpythonRepl):
         """Like request_refresh, but for undo request events."""
         raise NotImplementedError
 
-    def on_suspend():
+    def on_suspend(self):
         """Will be called on sigtstp.
 
         Do whatever cleanup would allow the user to use other programs."""
         raise NotImplementedError
 
-    def after_suspend():
+    def after_suspend(self):
         """Will be called when process foregrounded after suspend.
 
         See to it that process_event is called with None to trigger a refresh
@@ -528,8 +532,11 @@ class BaseRepl(BpythonRepl):
         sys.stdin = self.stdin
         self.orig_sigwinch_handler = signal.getsignal(signal.SIGWINCH)
         self.orig_sigtstp_handler = signal.getsignal(signal.SIGTSTP)
-        signal.signal(signal.SIGWINCH, self.sigwinch_handler)
-        signal.signal(signal.SIGTSTP, self.sigtstp_handler)
+
+        if is_main_thread():
+            # This turns off resize detection and ctrl-z suspension.
+            signal.signal(signal.SIGWINCH, self.sigwinch_handler)
+            signal.signal(signal.SIGTSTP, self.sigtstp_handler)
 
         self.orig_meta_path = sys.meta_path
         if self.watcher:
@@ -542,8 +549,12 @@ class BaseRepl(BpythonRepl):
         sys.stdin = self.orig_stdin
         sys.stdout = self.orig_stdout
         sys.stderr = self.orig_stderr
-        signal.signal(signal.SIGWINCH, self.orig_sigwinch_handler)
-        signal.signal(signal.SIGTSTP, self.orig_sigtstp_handler)
+
+        if is_main_thread():
+            # This turns off resize detection and ctrl-z suspension.
+            signal.signal(signal.SIGWINCH, self.orig_sigwinch_handler)
+            signal.signal(signal.SIGTSTP, self.orig_sigtstp_handler)
+
         sys.meta_path = self.orig_meta_path
 
     def sigwinch_handler(self, signum, frame):
@@ -675,6 +686,8 @@ class BaseRepl(BpythonRepl):
             self.down_one_line()
         elif e in ("<Ctrl-d>",):
             self.on_control_d()
+        elif e in ("<Ctrl-o>",):
+            self.operate_and_get_next()
         elif e in ("<Esc+.>",):
             self.get_last_word()
         elif e in key_dispatch[self.config.reverse_incremental_search_key]:
@@ -786,9 +799,11 @@ class BaseRepl(BpythonRepl):
         else:
             self.cut_buffer = cut
 
-    def on_enter(self, insert_into_history=True):
+    def on_enter(self, insert_into_history=True, reset_rl_history=True):
         # so the cursor isn't touching a paren TODO: necessary?
         self._set_cursor_offset(-1, update_completion=False)
+        if reset_rl_history:
+            self.rl_history.reset()
 
         self.history.append(self.current_line)
         self.push(self.current_line, insert_into_history=insert_into_history)
@@ -851,6 +866,11 @@ class BaseRepl(BpythonRepl):
     def yank_from_buffer(self):
         pass
 
+    def operate_and_get_next(self):
+        # If we have not navigated back in history
+        # ctrl+o will have the same effect as enter
+        self.on_enter(reset_rl_history=False)
+
     def up_one_line(self):
         self.rl_history.enter(self.current_line)
         self._set_current_line(tabs_to_spaces(self.rl_history.back(
@@ -1000,7 +1020,7 @@ class BaseRepl(BpythonRepl):
         # or disappear; whenever current line or cursor offset changes,
         # unless this happened via selecting a match
         self.current_match = None
-        self.list_win_visible = BpythonRepl.complete(self, tab)
+        self.list_win_visible = self.complete(tab)
 
     def predicted_indent(self, line):
         # TODO get rid of this! It's repeated code! Combine with Repl.
@@ -1069,14 +1089,11 @@ class BaseRepl(BpythonRepl):
             if err:
                 indent = 0
 
-            # TODO This should be printed ABOVE the error that just happened
-            # instead or maybe just thrown away and not shown
-            if self.current_stdouterr_line:
-                self.display_lines.extend(paint.display_linize(
-                    self.current_stdouterr_line, self.width))
-                self.current_stdouterr_line = ''
-
-            self._set_current_line(' ' * indent, update_completion=True)
+            if self.rl_history.index == 0:
+                self._set_current_line(' ' * indent, update_completion=True)
+            else:
+                self._set_current_line(self.rl_history.entries[-self.rl_history.index],
+                                       reset_rl_history=False)
             self.cursor_offset = len(self.current_line)
 
     def keyboard_interrupt(self):
@@ -1123,8 +1140,13 @@ class BaseRepl(BpythonRepl):
     def get_current_block(self):
         return '\n'.join(self.buffer + [self.current_line])
 
-    def send_to_stdout(self, output):
-        """Send unicode string to Repl stdout"""
+    def send_to_stdouterr(self, output):
+        """Send unicode strings or FmtStr to Repl stdout or stderr
+
+        Must be able to handle FmtStrs because interpreter pass in
+        tracebacks already formatted."""
+        if not output:
+            return
         lines = output.split('\n')
         logger.debug('display_lines: %r', self.display_lines)
         self.current_stdouterr_line += lines[0]
@@ -1138,18 +1160,6 @@ class BaseRepl(BpythonRepl):
             self.current_stdouterr_line = lines[-1]
         logger.debug('display_lines: %r', self.display_lines)
 
-    def send_to_stderr(self, error):
-        """Send unicode strings or FmtStr to Repl stderr
-
-        Must be able to handle FmtStrs because interpreter pass in
-        tracebacks already formatted."""
-        lines = error.split('\n')
-        if lines[-1]:
-            self.current_stdouterr_line += lines[-1]
-        self.display_lines.extend(sum((paint.display_linize(line, self.width,
-                                                            blank_line=True)
-                                       for line in lines[:-1]), []))
-
     def send_to_stdin(self, line):
         if line.endswith('\n'):
             self.display_lines.extend(
@@ -1612,7 +1622,7 @@ class BaseRepl(BpythonRepl):
 
         if not self.weak_rewind:
             self.interp = self.interp.__class__()
-            self.interp.write = self.send_to_stderr
+            self.interp.write = self.send_to_stdouterr
             self.coderunner.interp = self.interp
             self.initialize_interp()
 
@@ -1702,7 +1712,7 @@ class BaseRepl(BpythonRepl):
         try:
             source = self.get_source_of_current_name()
         except SourceNotFound as e:
-            self.status_bar.message(str(e))
+            self.status_bar.message('%s' % (e, ))
         else:
             if self.config.highlight_show_source:
                 source = format(PythonLexer().get_tokens(source),
